from jnprsr import *
from sys import stdin

def pretty():
    print("[Type CTRL+D or '!END' at a new line to end input]")
    input_data = ""
    for line in stdin:
        if line.startswith("!END"):
            break
        input_data += line
    # We simply generate an abstract syntax tree
    ast = get_ast(input_data)
    # ... and rendering the config again based on that!
    out = render_config_from_ast(ast)
    print(out)



# Generated from JuniperParser.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from jnprsr.parser.JuniperParser import JuniperParser
else:
    from jnprsr.parser.JuniperParser import JuniperParser


# This class defines a complete listener for a parse tree produced by JuniperParser.
class JuniperParserListener(ParseTreeListener):

    # Enter a parse tree produced by JuniperParser#braced_clause.
    def enterBraced_clause(self, ctx:JuniperParser.Braced_clauseContext):
        pass

    # Exit a parse tree produced by JuniperParser#braced_clause.
    def exitBraced_clause(self, ctx:JuniperParser.Braced_clauseContext):
        pass


    # Enter a parse tree produced by JuniperParser#bracketed_clause.
    def enterBracketed_clause(self, ctx:JuniperParser.Bracketed_clauseContext):
        pass

    # Exit a parse tree produced by JuniperParser#bracketed_clause.
    def exitBracketed_clause(self, ctx:JuniperParser.Bracketed_clauseContext):
        pass


    # Enter a parse tree produced by JuniperParser#juniper_configuration.
    def enterJuniper_configuration(self, ctx:JuniperParser.Juniper_configurationContext):
        pass

    # Exit a parse tree produced by JuniperParser#juniper_configuration.
    def exitJuniper_configuration(self, ctx:JuniperParser.Juniper_configurationContext):
        pass


    # Enter a parse tree produced by JuniperParser#statement.
    def enterStatement(self, ctx:JuniperParser.StatementContext):
        pass

    # Exit a parse tree produced by JuniperParser#statement.
    def exitStatement(self, ctx:JuniperParser.StatementContext):
        pass


    # Enter a parse tree produced by JuniperParser#terminator.
    def enterTerminator(self, ctx:JuniperParser.TerminatorContext):
        pass

    # Exit a parse tree produced by JuniperParser#terminator.
    def exitTerminator(self, ctx:JuniperParser.TerminatorContext):
        pass


    # Enter a parse tree produced by JuniperParser#word.
    def enterWord(self, ctx:JuniperParser.WordContext):
        pass

    # Exit a parse tree produced by JuniperParser#word.
    def exitWord(self, ctx:JuniperParser.WordContext):
        pass



del JuniperParser


from sys import stdin
from jnprsr import *
import prompt_toolkit
from jnprsr.prompt_toolkit_custom_nested_completer import CustomNestedCompleter
import pprint
from anytree.resolver import ChildResolverError

def subtree():
    print("[Type CTRL+D or '!END' at a new line to end input]")
    input_data = ""
    for line in stdin:
        if line.startswith("!END"):
            break
        input_data += line


    ast = get_ast(input_data)
    session = prompt_toolkit.PromptSession()

    # Transform AST into dict for completer
    completer_dict = {
        "show": {
            "configuration":
                render_dict_from_ast(ast)["root"]
        }
    }

    # Instantiate completer from dict
    completer = CustomNestedCompleter.from_nested_dict(completer_dict)

    # Setup Command Prompt
    while True:
        path = session.prompt("jnprsr> ", completer=completer)
        path = path.replace("show configuration ", "")
        try:
            subast = get_sub_tree(ast, path)
            print(render_config_from_ast(subast))
        except ChildResolverError as e:
            print("Error", e)






"""jnprsr is a Parser for Juniper Configuration Files"""

__version__ = "0.0.7"
__author__ = "markusju"
__author_email__ = "markus.jungbluth@gmail.com"

from .utils import get_ast, render_config_from_ast, render_ascii_tree_from_ast, merge, get_sub_tree, render_dict_from_ast


class JuniperASTRenderer():
    def __init__(self):
        self.indent = "    "
        self.buffer = ""

    def render(self, node):
        self.buffer = ""
        self.__render(node, level=-1)
        return self.buffer

    def __render(self, node, level):
        # Skip artificial root node
        if level > -1:
            self.buffer += self.indent * level + node.name
            if node.children:
                self.buffer += " {\n"
            else:
                self.buffer += ";\n"
        # Recursion!
        children = [self.__render(child, level=level+1) for child in node.children]
        # Skip closing bracket of artificial root node
        if children and level > -1:
            self.buffer += self.indent * level + "}\n"


# Generated from JuniperParser.g4 by ANTLR 4.10.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,13,56,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,1,0,1,
        0,5,0,15,8,0,10,0,12,0,18,9,0,1,0,1,0,1,1,1,1,4,1,24,8,1,11,1,12,
        1,25,1,1,1,1,1,2,4,2,31,8,2,11,2,12,2,32,1,2,1,2,1,3,3,3,38,8,3,
        1,3,4,3,41,8,3,11,3,12,3,42,1,3,1,3,1,3,1,3,1,3,3,3,50,8,3,1,4,1,
        4,1,5,1,5,1,5,0,0,6,0,2,4,6,8,10,0,1,2,0,1,1,5,5,56,0,12,1,0,0,0,
        2,21,1,0,0,0,4,30,1,0,0,0,6,37,1,0,0,0,8,51,1,0,0,0,10,53,1,0,0,
        0,12,16,5,8,0,0,13,15,3,6,3,0,14,13,1,0,0,0,15,18,1,0,0,0,16,14,
        1,0,0,0,16,17,1,0,0,0,17,19,1,0,0,0,18,16,1,0,0,0,19,20,5,2,0,0,
        20,1,1,0,0,0,21,23,5,9,0,0,22,24,3,10,5,0,23,22,1,0,0,0,24,25,1,
        0,0,0,25,23,1,0,0,0,25,26,1,0,0,0,26,27,1,0,0,0,27,28,5,3,0,0,28,
        3,1,0,0,0,29,31,3,6,3,0,30,29,1,0,0,0,31,32,1,0,0,0,32,30,1,0,0,
        0,32,33,1,0,0,0,33,34,1,0,0,0,34,35,5,0,0,1,35,5,1,0,0,0,36,38,7,
        0,0,0,37,36,1,0,0,0,37,38,1,0,0,0,38,40,1,0,0,0,39,41,3,10,5,0,40,
        39,1,0,0,0,41,42,1,0,0,0,42,40,1,0,0,0,42,43,1,0,0,0,43,49,1,0,0,
        0,44,50,3,0,0,0,45,46,3,2,1,0,46,47,3,8,4,0,47,50,1,0,0,0,48,50,
        3,8,4,0,49,44,1,0,0,0,49,45,1,0,0,0,49,48,1,0,0,0,50,7,1,0,0,0,51,
        52,5,11,0,0,52,9,1,0,0,0,53,54,5,12,0,0,54,11,1,0,0,0,6,16,25,32,
        37,42,49
    ]

class JuniperParser ( Parser ):

    grammarFileName = "JuniperParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'replace:'", "'}'", "']'", "')'", "'inactive:'", 
                     "<INVALID>", "<INVALID>", "'{'", "'['", "'('", "';'" ]

    symbolicNames = [ "<INVALID>", "REPLACE", "CLOSE_BRACE", "CLOSE_BRACKET", 
                      "CLOSE_PAREN", "INACTIVE", "LINE_COMMENT", "MULTILINE_COMMENT", 
                      "OPEN_BRACE", "OPEN_BRACKET", "OPEN_PAREN", "SEMICOLON", 
                      "WORD", "WS" ]

    RULE_braced_clause = 0
    RULE_bracketed_clause = 1
    RULE_juniper_configuration = 2
    RULE_statement = 3
    RULE_terminator = 4
    RULE_word = 5

    ruleNames =  [ "braced_clause", "bracketed_clause", "juniper_configuration", 
                   "statement", "terminator", "word" ]

    EOF = Token.EOF
    REPLACE=1
    CLOSE_BRACE=2
    CLOSE_BRACKET=3
    CLOSE_PAREN=4
    INACTIVE=5
    LINE_COMMENT=6
    MULTILINE_COMMENT=7
    OPEN_BRACE=8
    OPEN_BRACKET=9
    OPEN_PAREN=10
    SEMICOLON=11
    WORD=12
    WS=13

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.10.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Braced_clauseContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACE(self):
            return self.getToken(JuniperParser.OPEN_BRACE, 0)

        def CLOSE_BRACE(self):
            return self.getToken(JuniperParser.CLOSE_BRACE, 0)

        def statement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JuniperParser.StatementContext)
            else:
                return self.getTypedRuleContext(JuniperParser.StatementContext,i)


        def getRuleIndex(self):
            return JuniperParser.RULE_braced_clause

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBraced_clause" ):
                listener.enterBraced_clause(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBraced_clause" ):
                listener.exitBraced_clause(self)




    def braced_clause(self):

        localctx = JuniperParser.Braced_clauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_braced_clause)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 12
            self.match(JuniperParser.OPEN_BRACE)
            self.state = 16
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << JuniperParser.REPLACE) | (1 << JuniperParser.INACTIVE) | (1 << JuniperParser.WORD))) != 0):
                self.state = 13
                self.statement()
                self.state = 18
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 19
            self.match(JuniperParser.CLOSE_BRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Bracketed_clauseContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACKET(self):
            return self.getToken(JuniperParser.OPEN_BRACKET, 0)

        def CLOSE_BRACKET(self):
            return self.getToken(JuniperParser.CLOSE_BRACKET, 0)

        def word(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JuniperParser.WordContext)
            else:
                return self.getTypedRuleContext(JuniperParser.WordContext,i)


        def getRuleIndex(self):
            return JuniperParser.RULE_bracketed_clause

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBracketed_clause" ):
                listener.enterBracketed_clause(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBracketed_clause" ):
                listener.exitBracketed_clause(self)




    def bracketed_clause(self):

        localctx = JuniperParser.Bracketed_clauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_bracketed_clause)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 21
            self.match(JuniperParser.OPEN_BRACKET)
            self.state = 23 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 22
                self.word()
                self.state = 25 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==JuniperParser.WORD):
                    break

            self.state = 27
            self.match(JuniperParser.CLOSE_BRACKET)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Juniper_configurationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(JuniperParser.EOF, 0)

        def statement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JuniperParser.StatementContext)
            else:
                return self.getTypedRuleContext(JuniperParser.StatementContext,i)


        def getRuleIndex(self):
            return JuniperParser.RULE_juniper_configuration

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterJuniper_configuration" ):
                listener.enterJuniper_configuration(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitJuniper_configuration" ):
                listener.exitJuniper_configuration(self)




    def juniper_configuration(self):

        localctx = JuniperParser.Juniper_configurationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_juniper_configuration)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 30 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 29
                self.statement()
                self.state = 32 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << JuniperParser.REPLACE) | (1 << JuniperParser.INACTIVE) | (1 << JuniperParser.WORD))) != 0)):
                    break

            self.state = 34
            self.match(JuniperParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class StatementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self._word = None # WordContext
            self.words = list() # of WordContexts

        def braced_clause(self):
            return self.getTypedRuleContext(JuniperParser.Braced_clauseContext,0)


        def terminator(self):
            return self.getTypedRuleContext(JuniperParser.TerminatorContext,0)


        def INACTIVE(self):
            return self.getToken(JuniperParser.INACTIVE, 0)

        def REPLACE(self):
            return self.getToken(JuniperParser.REPLACE, 0)

        def word(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JuniperParser.WordContext)
            else:
                return self.getTypedRuleContext(JuniperParser.WordContext,i)


        def bracketed_clause(self):
            return self.getTypedRuleContext(JuniperParser.Bracketed_clauseContext,0)


        def getRuleIndex(self):
            return JuniperParser.RULE_statement

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStatement" ):
                listener.enterStatement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStatement" ):
                listener.exitStatement(self)




    def statement(self):

        localctx = JuniperParser.StatementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_statement)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 37
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==JuniperParser.REPLACE or _la==JuniperParser.INACTIVE:
                self.state = 36
                _la = self._input.LA(1)
                if not(_la==JuniperParser.REPLACE or _la==JuniperParser.INACTIVE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


            self.state = 40 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 39
                localctx._word = self.word()
                localctx.words.append(localctx._word)
                self.state = 42 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==JuniperParser.WORD):
                    break

            self.state = 49
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [JuniperParser.OPEN_BRACE]:
                self.state = 44
                self.braced_clause()
                pass
            elif token in [JuniperParser.OPEN_BRACKET]:
                self.state = 45
                self.bracketed_clause()
                self.state = 46
                self.terminator()
                pass
            elif token in [JuniperParser.SEMICOLON]:
                self.state = 48
                self.terminator()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class TerminatorContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def SEMICOLON(self):
            return self.getToken(JuniperParser.SEMICOLON, 0)

        def getRuleIndex(self):
            return JuniperParser.RULE_terminator

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterTerminator" ):
                listener.enterTerminator(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitTerminator" ):
                listener.exitTerminator(self)




    def terminator(self):

        localctx = JuniperParser.TerminatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_terminator)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 51
            self.match(JuniperParser.SEMICOLON)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class WordContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WORD(self):
            return self.getToken(JuniperParser.WORD, 0)

        def getRuleIndex(self):
            return JuniperParser.RULE_word

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterWord" ):
                listener.enterWord(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitWord" ):
                listener.exitWord(self)




    def word(self):

        localctx = JuniperParser.WordContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_word)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 53
            self.match(JuniperParser.WORD)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from jnprsr.parser.JuniperParserListener import JuniperParserListener
from jnprsr.parser.JuniperParser import JuniperParser
from anytree import NodeMixin
from anytree.node.util import _repr

class JuniperASTNode(NodeMixin, object):
    def __init__(self, name, parent=None):
        self.name = name
        self.key = ""
        self.value = ""
        self.parent = parent
        self.separator = " "
        super(JuniperASTNode, self).__init__()

    def is_juniper_bracketed_clause(self):
        return "[" in self.name and "]" in self.name

    def is_juniper_leaf_node(self):
        return self.is_leaf

    def is_juniper_container_node(self):
        return not self.is_leaf

    def append_value(self, value):
        if not self.is_juniper_bracketed_clause():
            raise ValueError("Can only be used on bracketed clauses")
        self.value += value
        self.name = self.key + " [" + self.value + " ]"

    def __repr__(self):
        args = ["%r" % " ".join([""] + [str(node.name) for node in self.path])]
        return _repr(self, args=args, nameblacklist=["name", "value", "key"])

class JuniperAST(JuniperParserListener):
    """
    This Class inherits from the JuniperParserListener created by ANTLR4.
    We are hooking onto various methods called during tree traversal to print out the config.
    A buffer used inside the class collects the configuration tree reconstructed from the the tree.
    """
    def __init__(self):
        self.root = JuniperASTNode("root")
        self.point = self.root
        self.depth = 0
        self.in_bracketed_clause = False

        # Enter a parse tree produced by JuniperParser#braced_clause.
    def enterBraced_clause(self, ctx: JuniperParser.Braced_clauseContext):
        pass

        # Exit a parse tree produced by JuniperParser#braced_clause.

    def exitBraced_clause(self, ctx: JuniperParser.Braced_clauseContext):
        #pass
        self.point = self.point.parent

        # Enter a parse tree produced by JuniperParser#bracketed_clause.

    def enterBracketed_clause(self, ctx: JuniperParser.Bracketed_clauseContext):
        self.point.name += " ["
        # Exit a parse tree produced by JuniperParser#bracketed_clause.

    def exitBracketed_clause(self, ctx: JuniperParser.Bracketed_clauseContext):
        self.point.name += " ]"
        # Enter a parse tree produced by JuniperParser#juniper_configuration.

    def enterJuniper_configuration(self, ctx: JuniperParser.Juniper_configurationContext):
        pass

        # Exit a parse tree produced by JuniperParser#juniper_configuration.

    def exitJuniper_configuration(self, ctx: JuniperParser.Juniper_configurationContext):
        pass

        # Enter a parse tree produced by JuniperParser#statement.

    def enterStatement(self, ctx: JuniperParser.StatementContext):
        node = JuniperASTNode(name="", parent=self.point)
        self.point = node
        # Exit a parse tree produced by JuniperParser#statement.

    def exitStatement(self, ctx: JuniperParser.StatementContext):
        pass
        # Enter a parse tree produced by JuniperParser#terminator.

    def enterTerminator(self, ctx: JuniperParser.TerminatorContext):
        pass
        # Exit a parse tree produced by JuniperParser#terminator.

    def exitTerminator(self, ctx: JuniperParser.TerminatorContext):
        self.point = self.point.parent
        # Enter a parse tree produced by JuniperParser#word.

    def enterWord(self, ctx: JuniperParser.WordContext):
        if self.in_bracketed_clause:
            node = JuniperASTNode(name="", parent=self.point)
            self.point = node

        if len(self.point.name) < 1:
            self.point.name += ctx.getText()
            self.point.key = ctx.getText()
        else:
            self.point.name += " " + ctx.getText()
            self.point.value += " " + ctx.getText()

        if self.in_bracketed_clause:
            self.point = self.point.parent
        # Exit a parse tree produced by JuniperParser#word.

    def exitWord(self, ctx: JuniperParser.WordContext):
        pass





import antlr4
from jnprsr.parser.JuniperParser import JuniperParser
from jnprsr.parser.JuniperLexer import JuniperLexer
from jnprsr.parser.JuniperAST import JuniperAST, JuniperASTNode
from jnprsr.parser.JuniperASTRenderer import JuniperASTRenderer
from anytree import RenderTree, Resolver
from jnprsr.anytree_custom_dict_exporter import CustomDictExporter

def get_ast(input_data: str) -> JuniperASTNode:
    # Get Raw Data
    raw_input = antlr4.InputStream(input_data)

    # Run through Lexer
    tok_stream = antlr4.CommonTokenStream(
        JuniperLexer(raw_input)
    )
    # Pass stream of Tokens to Parser
    parser = JuniperParser(tok_stream)

    # Parse Tree
    tree = parser.juniper_configuration()

    # Init AST Generator Listener
    listener = JuniperAST()

    # Walk Tree
    walker = antlr4.ParseTreeWalker()
    walker.walk(listener, tree)

    return listener.root

def render_config_from_ast(ast: JuniperASTNode) -> str:
    renderer = JuniperASTRenderer()
    return renderer.render(ast)

def render_ascii_tree_from_ast(ast: JuniperASTNode) -> str:
    buffer = ""
    for pre, _, node in RenderTree(ast):
        buffer += pre + node.name + "\n"
    return buffer

def render_dict_from_ast(ast: JuniperASTNode) -> dict:
    exporter = CustomDictExporter()
    return exporter.export(ast)

def __intersection(lst1, lst2):
    return list(set(lst1).intersection(lst2))

def __difference(lst1, lst2):
    return list(set(lst1).difference(lst2))

def merge(ast1, ast2):
    __merge(ast1, ast2)
    return ast1

def __merge(ast1: JuniperASTNode, ast2: JuniperASTNode):
    """
    Merges two JuniperASTs, ast2 is merged onto ast1. Similar to a "load merge" operation.
    :param ast1:
    :param ast2:
    :return:
    """
    EXCEPT_KEYS = ["interface"]
    print("merge", ast1, ast2)
    # Get Names of all the children at current level
    ast1_children = list(map(lambda x:x.name, ast1.children))
    ast2_children = list(map(lambda x:x.name, ast2.children))
    # Find common children
    common_children = __intersection(ast1_children, ast2_children)
    # Find missing children
    different_children = __difference(ast2_children, ast1_children)
    # Iterate over both lists for both trees

    # In case there are no children, just add them
    if len(ast1.children) == 0:
        for child_ast2 in ast2.children:
            child_ast2.parent = ast1

    # Follow paths and descend to common config items
    for child_ast1 in ast1.children:
        for child_ast2 in ast2.children:

            # If there is an identical path, go down -> Recursion
            if child_ast1.key == child_ast2.key and child_ast1.value == child_ast2.value and child_ast1.name in common_children:
                __merge(child_ast1, child_ast2)

            # If we are at the sub-tree where something needs to be added to AST1 for a pre-existing node, add it
            if child_ast2.name in different_children and child_ast1.key == child_ast2.key:
                # Make ast2 node parent reference the reference of ast1 node
                child_ast2.parent = child_ast1.parent

                # Remove from list of different children
                different_children.remove(child_ast2.name)

                # If it is a leaf node and not a bracketed clause and its key is not on the exception list, replace it!
                if child_ast1.is_juniper_leaf_node() and not child_ast1.is_juniper_bracketed_clause() and child_ast1.key not in EXCEPT_KEYS:
                    # Make child_ast1 node an orphan
                    child_ast1.parent = None

                # If it is a leaf node and a bracketed clause insert elements!
                if child_ast1.is_juniper_leaf_node() and child_ast1.is_juniper_bracketed_clause():
                   child_ast2.append_value(child_ast1.value)
                   # Make ast1 node an orphan
                   child_ast1.parent = None

    # Add non-existing items at this level to AST1 that have no pre-existing nodes on AST1
    for missing_children in different_children:
        for child_ast2 in ast2.children:
            if child_ast2.name == missing_children:
                # AST1 can simply adopt them
                child_ast2.parent = ast1



def get_sub_tree(ast: JuniperASTNode, path: str) -> JuniperASTNode:
    resolver = Resolver()
    return resolver.get(ast, path)